using System;

namespace EnigmaLib
{
    // include all features in installed enigma sim
    // some part needs to keep track of how many chars have been entered. if multiples
    //   are highlighted and deleted, we need to know how many times to call AutoReverse()
    //   for deleting all, just use Reset()

    public interface INamed
    {
        /// <summary>Get or set Name</summary>
        string Name { get; set; }
    }

    // ENCRYPTION PATTERN:
    // adjust for position (forward)
    // encrypt
    // adjust for position (backward)
    // send result to next rotor

    public class StaticRotor : INamed
    {
        private string _name;

        /// <summary>Ring setting (0-based): A-01 = 0, Z-26 = 25</summary>
        private int _ringSetting = 0;
        private bool[] _notches = new bool[26];

        /// <summary>From entry wheel to reflector</summary>
        private int[] _wiringIn;
        /// <summary>From reflector to entry wheel.
        /// Identical to wiringIn except reversed</summary>
        private int[] _wiringOut;

        public StaticRotor(string wiringSchema, params char[] notches)
        {
            if (wiringSchema.Length != 26)
                throw new Exception("Invalid wiring schema");

            if (notches != null)
                foreach (char c in notches)
                    _notches[StaticLib.CharToInt26(c)] = true;

            _wiringIn = StaticLib.GetInt26Array();
            _wiringOut = StaticLib.GetInt26Array();

            for (int i = 0; i < wiringSchema.Length; i++)
            {
                int ord = StaticLib.CharToInt26(wiringSchema[i]);

                // ensure slot hasn't been set before
                if (_wiringOut[ord] != -1)
                    throw new Exception("Invalid wiring schema. This slot "
                        + "has already been assigned to.");

                _wiringIn[i] = ord;
                _wiringOut[ord] = i;
            }
        }

        public string Name
        {
            get { return _name; }
            set { _name = value; }
        }

        public void SetRing(int ringSetting)
        {
            // if the same, nothing to do
            if (ringSetting == _ringSetting)
                return;

            _ringSetting = ringSetting;
            OnRingSettingsChanged();
        }
        public void SetRing(char ringSetting)
        {
            SetRing(StaticLib.CharToInt26(ringSetting));
        }
        public bool IsNotch(int pos)
        {
            return _notches[pos];
        }

        /// <summary>From entry wheel to reflector</summary>
        /// <param name="pos"></param>
        /// <returns></returns>
        public int EncryptIn(int position)
        {
            return encrypt(position, _wiringIn);
        }
        /// <summary>From reflector entry wheel</summary>
        /// <param name="pos"></param>
        /// <returns></returns>
        public int EncryptOut(int position)
        {
            return encrypt(position, _wiringOut);
        }

        private int encrypt(int position, int[] array)
        {
            if (position < 0 || position > 25)
                throw new Exception("Position must be between 0-25");

            int relativePos = position - _ringSetting;
            if (relativePos < 0)
                relativePos += 26;

            int relativeResult = array[relativePos];
            
            int result = relativeResult + _ringSetting;
            result = result % 26;
            
            return result;
        }

        #region Events
        public event EventHandler RingSettingsChanged;
        protected virtual void OnRingSettingsChanged()
        {
            if (RingSettingsChanged != null)
                RingSettingsChanged(this, new EventArgs());
        }
        #endregion
    }
    public abstract class EncryptionRotor : INamed
    {
        public char Encrypt(char pos)
        {
            int i = StaticLib.CharToInt26(pos);
            int e = Encrypt(i);
            char c = StaticLib.Int26ToChar(e);
            return c;
        }
        public abstract int Encrypt(int pos);
        public abstract string Name { get; set; }
    }
    public class Rotor : EncryptionRotor, IUiRotor
    {
        private StaticRotor _rotor;
        private int _startPosition;
        private int _position;

        private EncryptionRotor _nextRotor;
        /// <summary>Get current position.</summary>
        public int Position
        {
            get { return _position; }
            private set
            {
                // if the same, nothing to do
                if (value == _position)
                    return;

                _position = value;
                OnPositionChanged();
            }
        }
        /// <summary>Get start position. To set start position, use
        /// SetStartPosition(), ManualAdvance(), ManualReverse()</summary>
        public int StartPosition
        {
            get { return _startPosition; }
            private set
            {
                // if the same, nothing to do
                if (value == _startPosition)
                    return;

                _startPosition = value;
                OnStartPositionChanged();
            }
        }

        #region public c'tor
        public Rotor(StaticRotor rotor)
        {
            init(rotor);
        }
        public Rotor(string wiringSchema, params char[] notches)
        {
            StaticRotor rotor = new StaticRotor(wiringSchema, notches);
            init(rotor);
        }
        private void init(StaticRotor rotor)
        {
            if (rotor == null)
                throw new Exception("Must instantiate with a rotor");
            _rotor = rotor;
            Clear();
        }
        #endregion

        /// <summary>Set start position and reset position to it</summary>
        /// <param name="position"></param>
        public void SetStartPosition(int position)
        {
            StartPosition = position;
            Reset();
        }

        /// <summary>Revert to start position</summary>
        public void Reset()
        {
            Position = _startPosition;
        }

        public void Clear()
        {
            StartPosition = 0;
            Position = 0;
            _nextRotor = null;
        }

        public char GetVisibleLetter()
        {
            return StaticLib.Int26ToChar(_position);
        }

        public void SetNextRotor(EncryptionRotor nextRotor)
        {
            _nextRotor = nextRotor;
        }

        public override int Encrypt(int pos)
        {
            int relativePos = pos + _position;
            relativePos = relativePos % 26;

            int rotorIn = _rotor.EncryptIn(relativePos);

            int rotorInResult = rotorIn - _position;
            if (rotorInResult < 0)
                rotorInResult += 26;

            int nextRotorOut = _nextRotor.Encrypt(rotorInResult);

            int relativeNextRotorOut = nextRotorOut + _position;
            relativeNextRotorOut = relativeNextRotorOut % 26;

            int rotorOut = _rotor.EncryptOut(relativeNextRotorOut);

            int rotorOutResult = rotorOut - _position;
            if (rotorOutResult < 0)
                rotorOutResult += 26;

            return rotorOutResult;
        }

        /// <summary>Advance this rotor only. Ignore notches. Set starting position</summary>
        public void ManualAdvance()
        {
            advance();
            StartPosition = _position;
        }
        /// <summary>Reverse this rotor only. Ignore notches. Set starting position</summary>
        public void ManualReverse()
        {
            reverse();
            StartPosition = _position;
        }
        /// <summary>Advance Rotor. Advance notches if needed.</summary>
        public virtual void AutoAdvance()
        {
            advance();
        }
        /// <summary>Reverse rotor. Roll back notches if needed.
        /// Use for backspaces and clearing text</summary>
        public virtual void AutoReverse()
        {
            reverse();
        }

        private void advance()
        {
            Position = (_position + 1) % 26;
        }
        private void reverse()
        {
            int i = _position - 1;
            if (i < 0)
                i += 26;
            Position = i;
        }

        #region StaticRotor wrappers
        public override string Name
        {
            get { return _rotor.Name; }
            set { _rotor.Name = value; }
        }

        public void SetRing(char ringSetting)
        {
            _rotor.SetRing(ringSetting);
        }

        // needed for reversing
        public bool IsNotch(int pos)
        {
            pos = (_position + pos) % 26;
            return _rotor.IsNotch(pos);
        }

        // normally used
        public bool IsNotch()
        {
            return _rotor.IsNotch(_position);
        }
        #endregion

        #region Events
        public event EventHandler PositionChanged;
        public event EventHandler StartPositionChanged;
        protected virtual void OnPositionChanged()
        {
            if (PositionChanged != null)
                PositionChanged(this, new EventArgs());
        }
        protected virtual void OnStartPositionChanged()
        {
            if (StartPositionChanged != null)
                StartPositionChanged(this, new EventArgs());
        }
        public event EventHandler RingSettingsChanged
        {
            add { _rotor.RingSettingsChanged += value; }
            remove { _rotor.RingSettingsChanged -= value; }
        }
        #endregion
    }
    public class ThinRotor : Rotor
    {
        #region public c'tor
        public ThinRotor(StaticRotor rotor) : base(rotor) { }
        public ThinRotor(string wiringSchema, params char[] notches)
            : base(wiringSchema, notches) { }
        #endregion

        public override void AutoAdvance() { }
        public override void AutoReverse() { }
    }
    public class Reflector : EncryptionRotor
    {
        private string _name;
        /// <summary>Bidirectional (reflective/symetric) wiring</summary>
        private int[] _wiring;

        public Reflector(string wiringSchema)
        {
            if (wiringSchema.Length != 26)
                throw new Exception("Invalid wiring schema");

            _wiring = StaticLib.GetInt26Array();

            // test and set wiring
            for (int i = 0; i < wiringSchema.Length; i++)
            {
                int ord = StaticLib.CharToInt26(wiringSchema[i]);

                // ensure this slot hasn't been set before.
                // if it was, then ensure it was set reflectively
                if (_wiring[ord] != -1 && _wiring[ord] != i)
                    throw new Exception("Invalid wiring schema. Not reflective.");

                _wiring[i] = ord;
            }
        }

        public override int Encrypt(int pos)
        {
            return _wiring[pos];
        }

        public override string Name
        {
            get { return _name; }
            set { _name = value; }
        }
    }
    public class Plugboard : IUiPlugboard
    {
        /// <summary>Bidirectional (reflective/symetric) wiring</summary>
        private int[] _wiring = new int[26];
        private int _count = 0;
        public int Count
        {
            get { return _count; }
        }

        public Plugboard()
        {
            for (int i = 0; i < _wiring.Length; i++)
                _wiring[i] = i;
        }

        public int Encrypt(int position)
        {
            return _wiring[position];
        }

        /// <summary>Get plug pairs</summary>
        /// <returns></returns>
        public string[] GetPlugs()
        {
            string[] pairs = new string[_count];
            int curr = 0;
            for (int i = 0; i < _wiring.Length; i++)
                if (_wiring[i] < i)
                    pairs[curr++] = StaticLib.Int26ToChar(_wiring[i]).ToString() + StaticLib.Int26ToChar(i).ToString();
            return pairs;
        }
        public void SetPlugs(int plug1, int plug2)
        {
            if (_wiring[plug1] != plug1) //throw new Exception("Cannot swap. Space is occupied.");
                throw new Exception(
                    "Cannot put " + StaticLib.Int26ToChar(plug2).ToString() +
                    " in " + StaticLib.Int26ToChar(plug1).ToString() + ". " +
                    StaticLib.Int26ToChar(plug1).ToString() + " already links to " +
                    StaticLib.Int26ToChar(_wiring[plug1]).ToString() + "");

            if (_wiring[plug2] != plug2) //throw new Exception("Cannot swap. Space is occupied.");
                throw new Exception(
                    "Cannot put " + StaticLib.Int26ToChar(plug1).ToString() +
                    " in " + StaticLib.Int26ToChar(plug2).ToString() + ". " +
                    StaticLib.Int26ToChar(plug2).ToString() + " already links to " +
                    StaticLib.Int26ToChar(_wiring[plug2]).ToString() + "");
            _wiring[plug1] = plug2;
            _wiring[plug2] = plug1;

            _count++;

            OnPlugsAdded(StaticLib.Int26ToChar(plug1), StaticLib.Int26ToChar(plug2));
        }
        public void RemovePlugs(int plug1, int plug2)
        {
            if (_wiring[plug1] != plug2 || _wiring[plug2] != plug1)
                throw new Exception("Not a valid existing plug set.");
            _wiring[plug1] = plug1;
            _wiring[plug2] = plug2;

            _count--;

            OnPlugsRemoved(StaticLib.Int26ToChar(plug1), StaticLib.Int26ToChar(plug2));
        }
        public void ClearAllPlugs()
        {
            foreach (string plug in GetPlugs())
                RemovePlugs(plug[0], plug[1]);
        }

        #region Plugboard UI Members
        public void SetPlugs(char plug1, char plug2)
        {
            SetPlugs(StaticLib.CharToInt26(plug1), StaticLib.CharToInt26(plug2));
        }

        public void RemovePlugs(char plug1, char plug2)
        {
            RemovePlugs(StaticLib.CharToInt26(plug1), StaticLib.CharToInt26(plug2));
        }
        #endregion

        #region Events
        public event PlugsChangedEventHandler PlugsAdded;
        public event PlugsChangedEventHandler PlugsRemoved;
        protected virtual void OnPlugsAdded(char plug1, char plug2)
        {
            if (PlugsAdded != null)
                PlugsAdded(this, new PlugsChangedEventArgs(plug1, plug2));
        }
        protected virtual void OnPlugsRemoved(char plug1, char plug2)
        {
            if (PlugsRemoved != null)
                PlugsRemoved(this, new PlugsChangedEventArgs(plug1, plug2));
        }
        #endregion
    }

    public class PlugsChangedEventArgs : EventArgs
    {
        public char Plug1;
        public char Plug2;
        public PlugsChangedEventArgs(char plug1, char plug2)
        {
            Plug1 = plug1;
            Plug2 = plug2;
        }
    }
    public delegate void PlugsChangedEventHandler(object sender, PlugsChangedEventArgs e);
}
