using System;
using System.Collections.Generic;

namespace EnigmaLib
{
    public class EnigmaMachine : IUiEnigmaMachine
    {
        private Reflector _reflector;
        // count rotor numbers starting from the right, 0-based.
        // eg: reflector, 3, 2, 1, 0, entry wheel, plugboard
        private Rotor[] _rotors = new Rotor[1];
        private Plugboard _plugboard = new Plugboard();

        /// <summary>Use dictionary to make sure the same rotor instance
        /// is never used 2x. If a rotor were used more than once,
        /// 'next rotor' would yield unpredictable results</summary>
        private Dictionary<Rotor, int> _rotorList = new Dictionary<Rotor, int>();

        /// <summary>Normal = true. Set to false to enter cryptoanalysis mode.</summary>
        private bool _autoAdvance = true;

        public IUiPlugboard Plugboard
        {
            get { return _plugboard; }
        }

        public EnigmaMachine() { }
        public EnigmaMachine(int rotorCount)
        {
            SetRotorCount(rotorCount);
        }

        /// <summary>Normal = true. Set to false to enter cryptoanalysis mode.</summary>
        public bool AutoAdvance
        {
            get { return _autoAdvance; }
            set
            {
                // if the same, nothing to do
                if (value == _autoAdvance)
                    return;

                _autoAdvance = value;
                OnAutoAdvanceChanged();
            }
        }

        public int GetRotorCount()
        {
            return _rotors.Length;
        }
        public void SetRotorCount(int count)
        {
            if (count < 1)
                throw new Exception("Must have at least 1 rotor");

            // if the same, nothing to do
            if (count == _rotors.Length)
                return;

            Rotor[] newRotors = new Rotor[count];
            if (count < _rotors.Length)
            {
                // copy rotors
                for (int i = 0; i < count; i++)
                    newRotors[i] = _rotors[i];

                // final rotor's 'next rotor' is reflector
                if (newRotors[newRotors.Length - 1] != null)
                    newRotors[newRotors.Length - 1].SetNextRotor(_reflector);

                // reset 'next rotor' to null for all unused rotors
                for (int i = count; i < _rotors.Length; i++)
                    if (_rotors[i] != null)
                        _rotors[i].SetNextRotor(null);
            }
            else // count > _rotors.Length
            {
                // copy rotors
                for (int i = 0; i < _rotors.Length; i++)
                    newRotors[i] = _rotors[i];

                // final rotor's 'next rotor' is null
                if (newRotors[_rotors.Length - 1] != null)
                    newRotors[_rotors.Length - 1].SetNextRotor(null);
            }
            _rotors = newRotors;
            OnRotorCountChanged();
        }

        public void SetReflector(Reflector reflector)
        {
            // if the same, nothing to do
            if (reflector == _reflector)
                return;

            OnReflectorRemoved(_reflector);
            _reflector = reflector;
            OnReflectorAdded(_reflector);

            Rotor finalRotor = _rotors[_rotors.Length - 1];
            if (finalRotor != null)
                finalRotor.SetNextRotor(_reflector);
        }
        public void RemoveReflector()
        {
            SetReflector(null);
        }
        /// <summary>Get rotor. 1-based index</summary>
        /// <param name="position"></param>
        public Rotor GetRotor(int position)
        {
            return _rotors[position - 1];
        }
        /// <summary>Get the 1-based position of a rotor</summary>
        /// <param name="rotor"></param>
        /// <returns></returns>
        public int GetPosition(Rotor rotor)
        {
            return _rotorList[rotor] + 1;
        }
        /// <summary>Set rotor. 1-based index</summary>
        /// <param name="position"></param>
        /// <param name="rotor"></param>
        public void SetRotor(int position, Rotor rotor)
        {
            if (_rotorList.ContainsKey(rotor))
                throw new Exception("Cannot use the same rotor more than once.");

            // convert to 0-based index
            int index = position - 1;

            // do this 1st to make sure the position exists.
            // if the index is out of bounds, let .NET throw its own exception
            Rotor oldRotor = _rotors[index];

            // if the same, nothing to do
            if (rotor == oldRotor)
                return;

            // remove old rotor
            if (oldRotor != null)
            {
                oldRotor.Clear();
                _rotorList.Remove(oldRotor);
                // remember: index is 0-based and the event returns a 1-based position
                OnRotorRemoved(rotor, position);
            }

            // if not the 1st rotor, set prev rotor's 'next rotor' to new rotor
            if (index > 0)
            {
                Rotor prevRotor = _rotors[index - 1];
                if (prevRotor != null)
                    prevRotor.SetNextRotor(rotor);
            }

            // if new rotor is the last one, set its 'next rotor' to the reflector
            if (rotor != null)
            {
                if (index == _rotors.Length - 1)
                    rotor.SetNextRotor(_reflector);
                else // set new rotor's 'next rotor' to the next normal rotor
                    rotor.SetNextRotor(_rotors[index + 1]);

                _rotorList.Add(rotor, index);
            }

            // place new rotor into the set
            _rotors[index] = rotor;
            // remember: index is 0-based and the event returns a 1-based position
            OnRotorAdded(rotor, position);
        }
        public void RemoveRotor(Rotor rotor)
        {
            if (!_rotorList.ContainsKey(rotor))
                throw new Exception("Rotor is not here");

            int index = _rotorList[rotor];
            int position = index + 1;

            SetRotor(position, rotor);
        }

        public char Encrypt(char c)
        {
            // before stepping, make sure the entry is valid
            int cIn = StaticLib.CharToInt26(c);

            // REMEMBER: step 1st, then get encrypted letter. not the other way around!
            advance();

            cIn = _plugboard.Encrypt(cIn);
            cIn = _rotors[0].Encrypt(cIn);
            cIn = _plugboard.Encrypt(cIn);
            char cOut = StaticLib.Int26ToChar(cIn);
            return cOut;
        }

        /// <summary>Advance 1 step. Only works if AutoAdvance = true.</summary>
        private void advance()
        {
            if (!_autoAdvance)
                return;

            // must eval rotors in reverse order

            // rotors (other than the right-most) turn on their own notch or
            // on the notch of the rotor to their right. this results in
            // double-stepping
            for (int i = _rotors.Length - 1; i > 0; i--)
                if (_rotors[i].IsNotch() || _rotors[i - 1].IsNotch())
                    _rotors[i].AutoAdvance();

            // 1st rotor always turns
            if (_rotors[0] != null)
                _rotors[0].AutoAdvance();
        }

        /// <summary>Reverse 1 step. Only works if AutoAdvance = true.
        /// Note: reversing won't always result in the letters looking exactly like they
        /// did before the reverse, but since the machine is stepped before encryption,
        /// the encryption will still be correct</summary>
        public void Reverse()
        {
            if (!_autoAdvance)
                return;

            for (int i = _rotors.Length - 1; i > 0; i--)
                if (_rotors[i].IsNotch(25) || _rotors[i - 1].IsNotch(25))
                    _rotors[i].AutoReverse();

            _rotors[0].AutoReverse();
        }

        /// <summary>Reset to starting position</summary>
        public void Reset()
        {
            foreach (Rotor r in _rotors)
                r.Reset();
        }

        #region Events
        public event RotorChangedEventHandler RotorAdded;
        public event RotorChangedEventHandler RotorRemoved;
        public event ReflectorChangedEventHandler ReflectorAdded;
        public event ReflectorChangedEventHandler ReflectorRemoved;
        public event EventHandler AutoAdvanceChanged;
        public event EventHandler RotorCountChanged;

        protected virtual void OnRotorAdded(Rotor rotor, int position)
        {
            if (RotorAdded != null)
                RotorAdded(this, new RotorChangedEventArgs(rotor, position));
        }
        protected virtual void OnRotorRemoved(Rotor rotor, int position)
        {
            if (RotorRemoved != null)
                RotorRemoved(this, new RotorChangedEventArgs(rotor, position));
        }
        protected virtual void OnReflectorAdded(Reflector reflector)
        {
            if (ReflectorAdded != null)
                ReflectorAdded(this, new ReflectorChangedEventArgs(reflector));
        }
        protected virtual void OnReflectorRemoved(Reflector reflector)
        {
            if (ReflectorRemoved != null)
                ReflectorRemoved(this, new ReflectorChangedEventArgs(reflector));
        }
        protected virtual void OnAutoAdvanceChanged()
        {
            if (AutoAdvanceChanged != null)
                AutoAdvanceChanged(this, new EventArgs());
        }
        protected virtual void OnRotorCountChanged()
        {
            if (RotorCountChanged != null)
                RotorCountChanged(this, new EventArgs());
        }
        #endregion
    }

    public class RotorChangedEventArgs : EventArgs
    {
        public Rotor Rotor;
        /// <summary>1-based index</summary>
        public int Position;
        public RotorChangedEventArgs(Rotor rotor, int position)
        {
            Rotor = rotor;
            Position = position;
        }
    }
    public delegate void RotorChangedEventHandler(object sender, RotorChangedEventArgs e);

    public class ReflectorChangedEventArgs : EventArgs
    {
        public Reflector Reflector;
        public ReflectorChangedEventArgs(Reflector reflector)
        {
            Reflector = reflector;
        }
    }
    public delegate void ReflectorChangedEventHandler(object sender, ReflectorChangedEventArgs e);
}
