using System;
using System.Collections.Generic;

using EnigmaLib;

namespace HistoricalEnigma
{
    public class HistoricalEnigma : IUiHistoricalEnigma
    {
        private IUiEnigmaMachine _enigma = new EnigmaMachine(3);

        private Dictionary<HistoricalRotors, Rotor> _rotorDic
            = new Dictionary<HistoricalRotors, Rotor>();
        private Dictionary<HistoricalReflectors, Reflector> _reflectorDic
            = new Dictionary<HistoricalReflectors, Reflector>();

        private Dictionary<HistoricalEnigmaStates, HistoricalEnigmaState> _stateDic
            = new Dictionary<HistoricalEnigmaStates, HistoricalEnigmaState>();

        private List<Rotor> _availableRotors = new List<Rotor>();
        private List<Reflector> _availableReflectors = new List<Reflector>();

        private HistoricalEnigmaState _state;

        #region c'tor
        public HistoricalEnigma()
        {
            buildDictionaries();

            // set default state
            ChangeStateToM3Army();
        }

        private void buildDictionaries()
        {
            _rotorDic.Add(HistoricalRotors.I, HistoryFactory.RotorI());
            _rotorDic.Add(HistoricalRotors.II, HistoryFactory.RotorII());
            _rotorDic.Add(HistoricalRotors.III, HistoryFactory.RotorIII());
            _rotorDic.Add(HistoricalRotors.IV, HistoryFactory.RotorIV());
            _rotorDic.Add(HistoricalRotors.V, HistoryFactory.RotorV());
            _rotorDic.Add(HistoricalRotors.VI, HistoryFactory.RotorVI());
            _rotorDic.Add(HistoricalRotors.VII, HistoryFactory.RotorVII());
            _rotorDic.Add(HistoricalRotors.VIII, HistoryFactory.RotorVIII());
            _rotorDic.Add(HistoricalRotors.Beta, HistoryFactory.RotorBeta());
            _rotorDic.Add(HistoricalRotors.Gamma, HistoryFactory.RotorGamma());

            _reflectorDic.Add(HistoricalReflectors.A, HistoryFactory.ReflectorA());
            _reflectorDic.Add(HistoricalReflectors.B, HistoryFactory.ReflectorB());
            _reflectorDic.Add(HistoricalReflectors.C, HistoryFactory.ReflectorC());
            _reflectorDic.Add(HistoricalReflectors.BThin, HistoryFactory.ReflectorBThin());
            _reflectorDic.Add(HistoricalReflectors.CThin, HistoryFactory.ReflectorCThin());

            _stateDic.Add(HistoricalEnigmaStates.M3Army, new M3Army());
            _stateDic.Add(HistoricalEnigmaStates.M3Navy, new M3Navy());
            _stateDic.Add(HistoricalEnigmaStates.M4, new M4());
        }
        #endregion

        #region manage state
        public void ChangeStateToM3Army()
        {
            setState(_stateDic[HistoricalEnigmaStates.M3Army]);
        }
        public void ChangeStateToM3Navy()
        {
            setState(_stateDic[HistoricalEnigmaStates.M3Navy]);
        }
        public void ChangeStateToM4()
        {
            setState(_stateDic[HistoricalEnigmaStates.M4]);
        }
        private void setState(HistoricalEnigmaState state)
        {
            _state = state;
            OnStateChanged();
            _enigma.SetRotorCount(_state.GetQtyRotors());

            // change state. remove all rotors and reflectors not in the state's 'add' list
            foreach (KeyValuePair<HistoricalRotors, Rotor> kvp in _rotorDic)
            {
                if (_state.AvailRotors().Contains(kvp.Key))
                {
                    if (!_availableRotors.Contains(kvp.Value))
                        _availableRotors.Add(kvp.Value);
                }
                else
                {
                    if (_availableRotors.Contains(kvp.Value))
                    {
                        _availableRotors.Remove(kvp.Value);
                        _enigma.RemoveRotor(kvp.Value);
                    }
                }
            }
            foreach (KeyValuePair<HistoricalReflectors, Reflector> kvp in _reflectorDic)
            {
                if (_state.AvailReflectors().Contains(kvp.Key))
                {
                    if (!_availableReflectors.Contains(kvp.Value))
                        _availableReflectors.Add(kvp.Value);
                }
                else
                {
                    if (_availableReflectors.Contains(kvp.Value))
                    {
                        _availableReflectors.Remove(kvp.Value);
                        _enigma.RemoveReflector();
                    }
                }
            }
        }
        #endregion

        public bool Slot4IsVisible()
        {
            return _state.Slot4IsVisible();
        }

        #region set rotors and reflectors
        public void SetRotorInSlot1(Rotor rotor)
        {
            setThickRotor(1, rotor);
        }
        public void SetRotorInSlot2(Rotor rotor)
        {
            setThickRotor(2, rotor);
        }
        public void SetRotorInSlot3(Rotor rotor)
        {
            setThickRotor(3, rotor);
        }
        public void SetRotorInSlot4(ThinRotor rotor)
        {
            if (!_state.Slot4IsVisible())
                throw new Exception("Slot 4 is not currently available for use");
            setRotor(4, rotor);
        }

        private void setThickRotor(int position, Rotor rotor)
        {
            if (rotor is ThinRotor)
                throw new Exception("Only non-thin rotors can be used here.");

            setRotor(position, rotor);
        }
        private void setRotor(int position, Rotor rotor)
        {
            if (!_availableRotors.Contains(rotor))
                throw new Exception("This rotor is not in the "
                    + "availableRotors list. You cannot use it.");

            _enigma.SetRotor(position, rotor);
        }

        public void SetReflector(Reflector reflector)
        {
            if (!_availableReflectors.Contains(reflector))
                throw new Exception("This reflector is not in the "
                    + "availableReflectors list. You cannot use it.");

            _enigma.SetReflector(reflector);
        }
        #endregion

        #region Plugboard UI Members
        public int PlugPairCount
        {
            get { return _enigma.Plugboard.Count; }
        }

        /// <summary>Get plug pairs</summary>
        /// <returns></returns>
        public string[] GetPlugs()
        {
            return _enigma.Plugboard.GetPlugs();
        }

        public void SetPlugs(char plug1, char plug2)
        {
            if (_enigma.Plugboard.Count >= 11)
                throw new Exception("Historical enigma cannot exceed 11 plug pairs");

            _enigma.Plugboard.SetPlugs(plug1, plug2);
        }

        public void RemovePlugs(char plug1, char plug2)
        {
            _enigma.Plugboard.RemovePlugs(plug1, plug2);
        }
        #endregion

        #region Enigma UI Members
        /// <summary>Normal = true. Set to false to enter cryptoanalysis mode.</summary>
        public bool AutoAdvance
        {
            get { return _enigma.AutoAdvance; }
            set { _enigma.AutoAdvance = value; }
        }
        public char Encrypt(char c)
        {
            return _enigma.Encrypt(c);
        }
        /// <summary>Get the 1-based position of a rotor</summary>
        /// <param name="rotor"></param>
        /// <returns></returns>
        public int GetPosition(Rotor rotor)
        {
            return  _enigma.GetPosition(null);
        }
        /// <summary>Get rotor. 1-based index</summary>
        /// <param name="position"></param>
        public Rotor GetRotor(int position)
        {
            return _enigma.GetRotor(position);
        }
        public int GetRotorCount()
        {
            return _enigma.GetRotorCount();
        }
        public void RemoveRotor(Rotor rotor)
        {
            _enigma.RemoveRotor(rotor);
        }
        /// <summary>Reset to starting position</summary>
        public void Reset()
        {
            _enigma.Reset();
        }
        /// <summary>Reverse 1 step. Only works if AutoAdvance = true.
        /// Note: reversing won't always result in the letters looking exactly like they
        /// did before the reverse, but since the machine is stepped before encryption,
        /// the encryption will still be correct</summary>
        public void Reverse()
        {
            _enigma.Reverse();
        }
        public void RemoveReflector()
        {
            _enigma.RemoveReflector();
        }
        #endregion

        #region Events
        #region Wrapped Events
        public event RotorChangedEventHandler RotorAdded
        {
            add { _enigma.RotorAdded += value; }
            remove { _enigma.RotorAdded -= value; }
        }
        public event RotorChangedEventHandler RotorRemoved
        {
            add { _enigma.RotorRemoved += value; }
            remove { _enigma.RotorRemoved -= value; }
        }
        public event ReflectorChangedEventHandler ReflectorAdded
        {
            add { _enigma.ReflectorAdded += value; }
            remove { _enigma.ReflectorAdded -= value; }
        }
        public event ReflectorChangedEventHandler ReflectorRemoved
        {
            add { _enigma.ReflectorRemoved += value; }
            remove { _enigma.ReflectorRemoved -= value; }
        }
        public event EventHandler AutoAdvanceChanged
        {
            add { _enigma.AutoAdvanceChanged += value; }
            remove { _enigma.AutoAdvanceChanged -= value; }
        }
        public event EventHandler RotorCountChanged
        {
            add { _enigma.RotorCountChanged += value; }
            remove { _enigma.RotorCountChanged -= value; }
        }

        public event PlugsChangedEventHandler PlugsAdded
        {
            add { _enigma.Plugboard.PlugsAdded += value; }
            remove { _enigma.Plugboard.PlugsAdded -= value; }
        }
        public event PlugsChangedEventHandler PlugsRemoved
        {
            add { _enigma.Plugboard.PlugsRemoved += value; }
            remove { _enigma.Plugboard.PlugsRemoved -= value; }
        }
        #endregion

        public event EventHandler StateChanged;
        protected virtual void OnStateChanged()
        {
            if (StateChanged != null)
                StateChanged(this, new EventArgs());
        }
        #endregion
    }
}
